#
# $Source: /home/nlfm/Working/Mercator/Released/lib/RCS/Loader.tcl,v $
# $Date: 1999/12/15 13:24:36 $
# $Revision: 1.1.1.7 $
#
package provide Loader 1.0
#
package require Chart
package require Node
package require Link
#
proc loadChart {chart file} {
    if {[catch {open $file r} fd]} {
	tellError "Cannot open file \"$file\" : $fd"
	return 0
    }
    set buffer [read $fd]
    regsub -all {<([^> ]+)} $buffer "\} ; \\1 \{" buffer
    regsub -all {/?>} $buffer "\} \{" buffer
    close $fd
    namespace eval ::Loader "Start $chart {$buffer} ; End"
}
#
namespace eval ::Loader {
#
    variable map
    variable item
#
    proc parseArgs {str vars} {
	upvar 1 $vars values
	while {$str != {}} {
	    regexp " *(\[a-z\]+)=\"(\[^\"\]*)\"(.*)" $str m name v str
	    set values($name) $v
	}
    }
#
    proc Start {chart args} {
	variable map
	set map $chart
	catch {unset ::${map}::nameMap}
	array set ::${map}::nameMap {}
    }
#
    proc End {args} {
	variable map
	$map mapNames
    }
#
    proc ?xml {args} {
    }
#
    proc !DOCTYPE {args} {
    }
#
proc chart {pars args} {
    variable map
    parseArgs $pars values
    foreach {x y} [array get values] {
	switch $x last - start {
	    set ::${map}::$x $y
	} linktypes {
	    foreach dir [$map linkTypes] {
		unset ::${map}::linknames($dir)
		unset ::${map}::linkpairs($dir)
	    }
	    foreach {dir nm op} $y {
		set ::${map}::linknames($dir) $nm
		set ::${map}::linkpairs($dir) $op
	    }
	} default {
	}
    }
}
#
    proc /chart {args} {
    }
#
    proc node {pars args} {
        variable item
        variable map
        parseArgs $pars values
        set item [Node -map $map]
        foreach {x y} [array get values] {
	    switch -glob -- $x *links {
	        regexp {^(.*)links} $x m dir
	        foreach lk $y {
		    lappend ::${item}::links($dir) [$map nameMap($lk)]
	        }
	    } default {
	        if {[info exists ::${item}::$x]} {
	            set ::${item}::$x [unclean $y]
		} {
		    $item addUserField [list name $x] $y
		}
	    }
        }
        $map addnode $item
    }
#
    proc /node {args} {
        variable item
	set item {}
    }
#
    proc field {pars str} {
        variable item
	parseArgs $pars values
	$item addUserField [array get values] $str
    }
#
    proc /field {args} {
    }
#
    proc notes {pars str} {
	variable item
	regsub "^\n+" $str {} str
	regsub "\n+$" $str {} str
	switch "\n" $str { set str {} }
	set ::${item}::notes [unclean $str]
    }
#
    proc object {pars args} {
	variable item
	parseArgs $pars values
	if {[info exists values(name)]} {
	    switch -- $values(name) mail {
	       regsub {^(mailto:)*} [unclean $values(href)] {} ::${item}::mail
	    } default {
	        set ::${item}::$values(name) [unclean $values(href)]
	    }
	} {
	    foreach {x y} [array get values] {
	        switch $x mail {
		    regsub {^(mailto:)*} $y {} y
	        }
	        set ::${item}::$x $y
	    }
	}
    }
#
    proc /notes {args} {
    }
#
    proc link {pars args} {
	variable map
	variable item
	parseArgs $pars values
	set item [Link -map $map]
	foreach {x y} [array get values] {
	    if {[info exists ::${item}::$x]} {
	        set ::${item}::$x [unclean $y]
	    } {
	        $item addUserField [list name $x] $y
	    }
	}
	$map addlink $item
    }
#
    proc /link {args} {
        variable item
	set item {}
    }
#
    proc style {pars str} {
	variable map
	foreach {x y} $str {
	    $map addStyle $x $y
	}
    }
#
    proc /style {args} {
    }
#
    proc preferences {pars str} {
	variable map
	foreach {x y} [split $str ;] {
	    set ::${map}::[string trim $x] [string trim $y]
	}
    }
#
    proc /preferences {args} {
    }
#
# End of namespace Loader
#
}
