#
# $Source: /home/nlfm/Working/Mercator/Released/lib/RCS/Node.tcl,v $
# $Date: 1999/12/15 13:31:22 $
# $Revision: 1.1.1.25 $
#
package provide Node 1.0
#
package require Link
package require Chart
package require Object
package require Style
#
proc Node {args} {
    set this [::Object::makeObj Node $args]
    namespace eval ::$this {
	variable links
	variable map
	foreach x [$map linkTypes] {set links($x) {}}
    }
    return $this
}
#
namespace eval ::node {
    #
    proc configure {this args} {::Object::confObj $this $args}
    #
    proc fmtconf {this fmt} {
	set chrt [$this map]
	set w [$chrt canvas].$this
	switch {}\
	  [set val [set ::${this}::$fmt]] return
	if {[catch {
	    switch -- $fmt {
	    font {
		    switch $this\
		      [$chrt last] {
			    set ft [growfont $val]
			    $w.lb configure -font $ft
			}\
		      default {
			    $w.lb configure -font $val
			}
		}
	    foreground {
		    $w.lb configure -foreground $val
		}
	    background {
		    $w.lb configure -background $val
		    if {![listmember [$chrt selection] $this]} {
			$w configure -background $val
		    }
		}
	    }
	} msg]} {
	    tellError "Invalid $fmt \"$val\""
	    switch $fmt\
	      font {
		    switch [$chrt last]\
		      $this {
			    set ::${this}::$fmt [shrinkfont [$w.lb cget -$fmt]]
			}\
		      default {
			    set ::${this}::$fmt [$w.lb cget -$fmt]
			}
		}\
	      default {
		    set ::${this}::$fmt [$w.lb cget -$fmt]
		}
	    return 0
	}
	return 1
    }
    #
    proc delete {this} {
	[$this map] removeNode $this
	::Object::delete $this
    }
    #
    proc mapNames {this} {
	foreach dir [array names ::${this}::links] {
	    set tm {}
	    foreach lk [$this links($dir)] {
		lappend tm [[$this map] nameMap($lk)]
	    }
	    set ::${this}::links($dir) $tm
	}
    }
    #
    proc getname {this} {
	switch {}\
	  [set nm [$this name]] {
		set nm [set ::${this}::name $this]
	    }
	return $nm
    }
    #
    proc save {this desc} {
	global Node
	array set defs $Node
	puts -nonewline $desc "<node name=\"[$this getname]\" label=[quote\
	  [$this label]]"
	foreach x {style font foreground background keywords access deleted} {
	    switch -- [set ::${this}::$x]\
	      $defs($x) continue
	    puts -nonewline $desc "\n $x=[quote [set ::${this}::$x]]"
	}
	foreach {x y} [array get ::${this}::links] {
	    switch {}\
	      $y continue
	    set tls {}
	    foreach lk $y { lappend tls [$lk getname] }
	    puts -nonewline $desc "\n ${x}links=\"$tls\""
	}
	puts $desc >
	saveData $this $desc
	puts $desc </node>
    }
    #
    proc makeWindow {this xpos ypos anchor ft fg bg cv} {
	global prog
	trunctxt $this
	uplevel #0 trace variable ::${this}::label w "{trunctxt $this}"
	catch {::destroy $cv.$this}
	set fr [frame $cv.$this -relief raised -bd 2 -bg $bg]
	label $fr.lb -fg $fg -anchor center -bg $bg -font $ft -bd 0\
	  -relief flat -highlightthickness 0
	switch {}\
	  [set ::${this}::image] {
		$fr.lb configure -textvariable ::${this}::shorttext -width 15
	    }\
	  default {
		if {![set ::[$this map]::showimages] || {} == [set im\
		  [makeImage [set ::${this}::image]]]} {
		    $fr.lb configure -textvariable ::${this}::shorttext\
		      -width 15
		} {
		    $fr.lb configure -image $im
		}
	    }
	grid $fr.lb -sticky nsew -column 2 -row 0 -rowspan 2
	bind $fr.lb <Double-1> "$this open"
	bind $fr.lb <1> "$this select 0"
	bind $fr.lb <Shift-1> "$this select 1"
	bind $fr.lb <Control-1> "$this centre"
	bind $fr.lb <$prog(popmenu)> "$this menuNode %X %Y ; break"
	$cv create window $xpos $ypos -window $fr -tag $this -anchor $anchor
	update idletasks
    }
    #
    proc show {this level args} {
	set chrt [$this map]
	eval $this show[$chrt view] $chrt $level $args
    }
    #
    proc showoutline {this chrt level args} {
	set txt [$chrt text]
	switch $level\
	  1 {
		grid remove [$chrt canvas]
		grid $txt
		$txt delete 1.0 end
		$chrt tinsert end [$this label] [list $this level1] \n
		foreach lk [$this links(s)] { [$lk otherEnd $this] show 2 }
	    }\
	  default {
		$chrt tinsert end [$this label] [list $this level$level] \n
		incr level
		foreach lk [$this links(s)] { [$lk otherEnd $this] show $level }
	    }
    }
    #
    proc showcentred {this chrt level args} {
	set cv [$chrt canvas]
	set ft [$this font]
	set fg [$this foreground]
	set bg [$this background]
	if {[$this deleted]} {
	    if {![$chrt showdeleted]} return
	    set bg [tkDarken $bg 200]
	    set fg [tkDarken $fg 200]
	}
	switch $level\
	  1 {
		grid remove [$chrt text]
		grid [$chrt canvas]
		$chrt addHistory $this
		set ft [growfont $ft]
		set xpos [expr {[$chrt width]/2}]
		set ypos [expr {[$chrt height]/2}]
	    }\
	  2 {
		set xpos [lindex $args 0]
		set ypos [lindex $args 1]
	    }
	$this makeWindow $xpos $ypos center $ft $fg $bg $cv
	foreach {x1 y1 x2 y2} [$cv bbox $this] break
	switch $level\
	  1 {
		$this select 0
		set subs {}
		foreach dir [[$this map] linkTypes] {
		    if {[$chrt display $dir]} {
			set subs [concat $subs [$this links($dir)]]
		    }
		}
		$chrt circshow $this $subs
		set sx [halfit $x1 $x2]
		$cv create rectangle [expr $sx-4] [expr {$y1-4}] [expr {$sx +\
		  4}] [expr {$y1+4}] -fill green -outline blue -tag "$this n"
		$cv create rectangle [expr $sx-4] [expr {$y2-4}] [expr {$sx +\
		  4}] [expr {$y2+4}] -fill green -outline blue -tag "$this s"
		set sy [halfit $y1 $y2]
		$cv create rectangle [expr {$x1 - 4}] [expr {$sy-4}] [expr\
		  {$x1 + 4}] [expr {$sy+4}] -fill yellow -outline red\
		  -tag "$this w"
		$cv create rectangle [expr {$x2 - 4}] [expr {$sy-4}] [expr\
		  {$x2 + 4}] [expr {$sy+4}] -fill yellow -outline red\
		  -tag "$this e"
	    }\
	  default {
		set sy $y1
		switch [llength [concat [$this links(w)] [$this links(e)]]]\
		  0 -\
		  1 {  }\
		  default {
			$cv create rectangle [expr {$x1 - 4}] $sy [expr {$x1 +\
			  4}] [expr {$sy+8}] -fill green -outline blue\
			  -tag $this
			incr sy 10
		    }
		switch [llength [$this links(s)]]\
		  0 -\
		  1 {  }\
		  default {
			$cv create rectangle [expr {$x1 - 4}] $sy [expr {$x1 +\
			  4}] [expr {$sy+8}] -fill yellow -outline red\
			  -tag $this
		    }
		return
		if {[$chrt showalllinks]} {
		    foreach dir [$chrt linkTypes] {
			foreach lk [$this links($dir)] {
			    if {[winfo exists $cv.[$lk otherEnd $this]]} {
				$lk show
			    }
			}
		    }
		}
	    }
    }
    #
    proc showtree {this chrt level args} {
	set cv [$chrt canvas]
	set ft [$this font]
	set fg [$this foreground]
	set bg [$this background]
	if {[$this deleted]} {
	    if {![$chrt showdeleted]} return
	    set bg [tkDarken $bg 200]
	    set fg [tkDarken $fg 200]
	}
	switch $level\
	  1 {
		grid remove [$chrt text]
		grid [$chrt canvas]
		$chrt addHistory $this
		set ft [growfont $ft]
		set xpos [expr {[$chrt width]/2}]
		set ypos [expr {[$chrt height]/2}]
		set anchor center
	    }\
	  2 {
		set xpos [lindex $args 0]
		set ypos [lindex $args 1]
		set anchor [lindex $args 2]
	    }
	$this makeWindow $xpos $ypos $anchor $ft $fg $bg $cv
	switch $level\
	  1 {
		foreach {x1 y1 x2 y2} [$cv bbox $this] break
		set sx [expr {$x1 +($x2 - $x1) /2}]
		set sy [expr {$y1 +($y2 - $y1) /2}]
		$cv create polygon [expr {$sx - 16}] $y1 $sx [expr {$y1-8}]\
		  [expr {$sx + 16}] $y1 -fill green -outline blue -tag\
		  "h$this n" -smooth 1
		$cv create polygon [expr {$sx - 16}] $y2 $sx [expr {$y2+8}]\
		  [expr {$sx + 16}] $y2 -fill green -outline blue -tag\
		  "h$this s" -smooth 1
		$cv create polygon $x2 [expr {$sy - 4}] [expr {$x2+8}] $sy $x2\
		  [expr {$sy + 4}] -fill yellow -outline red -tag "h$this e"\
		  -smooth 1
		$cv create polygon $x1 [expr {$sy - 4}] [expr {$x1-8}] $sy $x1\
		  [expr {$sy + 4}] -fill yellow -outline red -tag "$this w"\
		  -smooth 1
		$this select 0
		update idletasks
		foreach dir [[$this map] linkTypes] {
		    if {[$chrt display $dir]} { $chrt famShow $this $dir }
		}
	    }\
	  default {
		foreach {x1 y1 x2 y2} [$cv bbox $this] break
		set sy $y1
		switch [llength [concat [$this links(w)] [$this links(e)]]]\
		  0 - 1 {}\
		  default {
			$cv create rectangle [expr {$x1 - 4}] $sy [expr {$x1 +\
			  4}] [expr {$sy+8}] -fill green -outline green\
			  -tag $this
			incr sy 10
		    }
		switch [llength [$this links(s)]]\
		  0 - 1 {}\
		  default {
			$cv create rectangle [expr {$x1 - 4}] $sy [expr {$x1 +\
			  4}] [expr {$sy+8}] -fill yellow -outline red\
			  -tag $this
		    }
		return
		if {[$chrt showalllinks]} {
		    foreach dir [$chrt linkTypes] {
			foreach lk [$this links($dir)] {
			    if {[winfo exists $cv.[$lk otherEnd $this]]} {
				$lk show
			    }
			}
		    }
		}
	    }
    }
    #
    proc centre {this} {
	[set cv [[$this map] canvas]] delete all
	$this show 1
    }
    #
    proc togdel {this} {
	if {[$this deleted]} {
	} {
	    $this relink 0
	    grid remove [[$this map] window].info.f1.del
	}
    }
    #
    proc open {this} {
	set chrt [$this map]
	set w [[$this map] window]
	if {[$this deleted]} {
	    grid $w.info.f1.del
	    $w.info.f1.del configure -variable ::${this}::deleted\
	      -command "$this togdel"
	} {
	    grid remove $w.info.f1.del
	}
	$w.info.name configure -textvariable ::${this}::label
	upvar #0 lastPane($w) oldPane lastObj($w) oldObj
	if {[info exists oldPane]} {
	    uplevel #0 trace vdelete ::${oldObj}::label w "{$chrt used}"
	}
	uplevel #0 trace variable ::${this}::label w "{$chrt used}"
	foreach x [list Notes Fields Format Keywords] {
	    $this init$x [::Notebook::getNote $w.info.nb $x]
	}
	set tab Notes
	switch {} [$this notes] {
	    switch {} [$this keywords] {} default { set tab Keywords }
	    foreach x {image url mail chart} {
		switch {} [$this $x] {} default {
		    set tab Fields
		    break
		}
	    }
	}
	::Notebook::pickNote $w.info.nb $tab
    }
    #
    proc close {this} {
	set w [[$this map] window]
	foreach x [list Notes Fields Format Keywords] {
	    $this unconf$x [::Notebook::getNote $w.info.nb $x]
	}
    }
    #
    proc addlink {this dir link} {
	lappend ::${this}::links($dir) $link
    }
    #
    proc select {this extend} {
	set chrt [$this map]
	if {!$extend} { $chrt deselect }
	$chrt addSelection $this
    }
    #
    proc buildNode {this type x y} {
	set chrt [$this map]
	set nm [Node -map $chrt]
	$chrt addlink [set lnk [Link -type $type -map $chrt -start $this\
	  -end $nm -colour [$chrt ink] -width [$chrt thick]]]
	set anchor [$chrt otherway $type]
	$this addlink $type $lnk
	$nm addlink $anchor $lnk
	switch $type {
	    n {
		set ::${lnk}::start $nm
		set ::${lnk}::end $this
		set ::${lnk}::type s
		$chrt removeRoot $this
	    }
	}
	$nm show 2 $x $y $anchor
	$lnk show
	$chrt used
	$nm select 0
	$chrt addnode $nm
	return $nm
    }
    #
    proc endLink {this type cv x y} {
	grab release $cv
	foreach {x1 y1 x2 y2} [$cv coords rubber] break
	$cv delete rubber
	bind $cv <ButtonRelease-1> {}
	buildNode $this $type $x $y
    }
    #
    # procs for handling opening and closing of node info. initXXX is called\
      when
    # the mode is selected. unconfXXX before any other node is selected.
    #
    proc initNotes {this win} {
	$win.txt delete 1.0 end
	$win.txt insert end [$this notes]
    }
    #
    proc unconfNotes {this win} {
	switch "\n" [set nts [$win.txt get 1.0 end]] {
	    set nts {}
	}
	set ::${this}::notes $nts
    }
    #
    proc showUserFields {this} {
	if {[winfo exists [set w .uf$this]]} {
	    wm deiconify $w
	    raise $w
	    return
	}
	global ufta prog
	toplevel $w
	wm title $w "[mc Node] \"[$this label]\" - [mc User Defined Fields]"
	wm protocol $w WM_DELETE_WINDOW "$this closeUF $w"
	grid columnconfigure $w 0 -weight 1
	set rw 0
	set chart [$this map]
	foreach x [namespace children ::${this}] {
	    grid [set f [frame $w.f$rw]] -sticky ew -row $rw
	    $this insertField $f $x
	    incr rw
	}
	grid [frame $w.btn] -sticky ew
	button $w.btn.new -text [mc "New Field"] -width 12 -command\
	  "$this newField $w"
	button $w.btn.ok -text [mc Close] -command "$this closeUF $w" -width 12
	grid $w.btn.new $w.btn.ok
    }
    #
    proc insertField {this f ns} {
        global prog
	set chart [$this map]
	grid [label $f.name -text [namespace tail $ns]\
	  -anchor w] -row 0 -column 0 -sticky w 
	bind $f.name <$prog(popmenu)> "$this editField $f $ns %X %Y"
	switch [set type [getval $ns UF_type text]] {
	    text {
		::entry $f.value -width 40 -textvariable ${ns}::value\
		  -relief sunken -bd 2
		trapid $f.value $chart
	    }
	    checkbox {
		set on [getval $ns UF_on 1]
		set off [getval $ns UF_off 0]
		::checkbutton $f.value -variable ${ns}::value -onvalue $on\
		  -offvalue $off -command "$chart used"
	    }
	    radio {
		frame $f.value
		set i 0
		foreach y [getval $ns UF_values {}] {
		    grid [::radiobutton $f.value.$i -value $y -text $y\
		      -variable ${ns}::value -command "$chart used"]\
		      -row 0 -column $i
		    incr i
		}
	    }
	    textarea {
		set width [getval $ns UF_rows 40]
		set height [getval $ns UF_cols 10]
		::text $f.value -width $width -height $height
		$f.value insert end [getval $ns value {}]
		lappend ufta([winfo toplevel $f]) $ns $f.value
		trapid $f.value $chart
	    }
	    file {
		frame $f.value
		grid [entry $f.value.e -width 30 -textvariable ${ns}::value \
		  -relief sunken -bd 2] -row 0 -column 1 -sticky ew
		grid [button $f.value.b -text [mc browse] -command "fileBrowse ${ns}::value"] \
		  -row 0 -column 1
		trapid $f.value.e $chart
	    }
	    default {
		label $f.value -text "Bad field type -\"$type\"" -foreground red
	    }
	}
	grid $f.value -padx 5 -pady 5 -row 0 -column 1 -sticky ew
	grid columnconfigure $f 0 -weight 1
	grid columnconfigure $f 1 -weight 3
    }
    #
    proc newField {this w} {
	array set foo [grid info $w.btn]
	set rw $foo(-row)
	grid $w.btn -row [expr {$rw+1}]
	grid [set f [frame $w.f$rw]] -sticky ew -row $rw
	set name field$rw
	$this addUserField [list name $name type text] {}
	$this insertField $f ::${this}::$name
    }
    #
    proc editField {this frame ns x y} {
        set m [menu .$this[namespace tail $ns] -tearoff 0]
	$m add command -label [mc Rename]
	$m add cascade -label [mc {Change Type}] -menu $m.type
	$m add separator
	$m add command -label [mc Delete] -command "$this deleteField $frame $ns"
	menu $m.type -tearoff 0
	foreach type {Text Checkbox Radio Textarea File} {
	    $m.type add radiobutton -label [mc $type] \
	      -value [string tolower $type] -variable ${ns}::UF_type\
	      -command "$this changeField $frame $ns [set ${ns}::UF_type]" 
	}
	tk_popup $m $x $y
    }
    #
    proc changeName {this frame ns} {
        ::destroy .$this[namespace tail $ns]
	[$this map] used
    }
    #
    proc deleteField {this frame ns} {
        ::destroy .$this[namespace tail $ns]
        namespace delete $ns
	::destroy $frame
	[$this map] used
    }
    #
    proc changeField {this frame ns oldtype} {
        switch $oldtype {
	    text { set dl {} }
	    checkbox { set dl {on off} }
	    radio { set dl {values} }
	    textarea { set dl {rows cols} }
	    file { set dl {} }
	}
	foreach x $dl {
	    namespace eval $ns "unset UF_$x"
	}
        destroy $frame.value $frame.name
	$this insertField $frame $ns
        ::destroy .$this[namespace tail $ns]
	[$this map] used
    }
    #
    proc closeUF {this w} {
        if {[winfo exists $w]} {
	    global ufta
	    if {[info exists ufta($w)]} {
	        foreach {x y} $ufta($w) {
		    catch {set ${x}::value [$y get 1.0 end]}
		}
		unset ufta($w)
	    }
	    ::destroy $w
	}
    }
    #
    proc initFields {this win} {
	set cmd "{[$this map] used}"
	set rw -1
	foreach {x b} {image 1 url 0 mail 0 chart 1 object 1} {
	    incr rw 2
	    $win.$rw.e configure -textvariable ::${this}::$x
	    uplevel #0 trace variable ::${this}::$x w $cmd
	    if {$b} {
		$win.$rw.b configure -command "${x}Browse $this"
	    }
	}
	$win.uf configure -command "$this showUserFields"
    }
    #
    proc unconfFields {this win} {
	$this closeUF .uf$this
    }
    #
    proc setfmt {this win fmt} {
	if {[$this fmtconf $fmt]} {
	    $win configure -$fmt [$this $fmt]
	}
    }
    #
    proc initFormat {this win} {
	set cmd "{[$this map] used}"
	foreach x {font foreground background} {
	    $win.$x.e configure -textvariable ::${this}::$x -$x [$this $x]
	    bind $win.$x.e <FocusOut> "$this setfmt %W $x"
	    bind $win.$x.e <Return> "$this setfmt %W $x"
	    catch {bind $win.$x.e <KP_Enter> "$this setfmt %W $x"}
	    uplevel #0 trace variable ::${this}::$x w $cmd
	    $win.$x.b configure -command "${x}Select $this"
	}
    }
    #
    proc unconfFormat {this win} {
	set cmd "{[$this map] used}"
	foreach x {font foreground background} {
	    switch {}\
	      [set v [$this $x]] {  }\
	      default { $this fmtconf $x }
	}
    }
    #
    proc initKeywords {this win} {
	bind $win.f.ent <Return> "
            set val \[%W get\]
            if {\[listadd ::${this}::keywords \$val\]} {
	        $win.lb insert end \$val
	    }
	    %W delete 0 end
	"
	bind $win.lb <Delete> "
            foreach x \[%W curselection\] {
	        listdel ::${this}::keywords \[%W get \$x\]
	        %W delete \$x
	    }
        "
	$win.f.ent delete 0 end
	$win.lb delete 0 end
	foreach x [$this keywords] {
	    $win.lb insert end $x
	}
    }
    #
    proc unconfKeywords {this win} {
    }
    #
    proc initAccess {this win} {
	global prog access
	switch -- [$this access] {
	.* { set access([$this map]) public }
	{} { set access([$this map]) private }
	default {
		set access([$this map]) other
		$win.f.ent configure -state normal
		foreach x [$this access] {
		    $win.lb insert end $x
		}
		$win.f.new configure -foreground $prog(fgdef)
	    }
	}
    }
    #
    proc unconfAccess {this win} {
    }
    #
    proc setAccess {this kind} {
	switch $kind {
	public { $this configure -access .* }
	private { $this configure -access {} }
	}
    }
    #
    proc linkSel {this dir} {
	set sel [[set chrt [$this map]] stash]
	switch $sel\
	  {} -\
	  $this return
	set odir [$chrt otherway $dir]
	set lnk [Link -map $chrt -type $dir -start $this -end $sel\
	  -colour [$chrt ink]]
	switch $dir {
	n {
		set ::${lnk}::start $nm
		set ::${lnk}::end $this
		set ::${lnk}::type $odir
		$chrt removeRoot $this
	    }
	}
	$this addlink $dir $lnk
	$sel addlink $odir $lnk
	$lnk show
    }
    #
    proc gourl {this url} {
	if {[catch {exec netscape -remote openurl($url,newwindow)}]} {
	    exec netscape $url &
	}
    }
    #
    proc gomail {this val} {
	tellInfo "Sorry not yet supported!"
    }
    #
    proc gochart {this file} {
	[Chart -file $file -name [basename $file]] show
    }
    #
    proc go {this} {
	foreach x {url mail chart} {
	    switch {}\
	      [set val [$this $x]] {  }\
	      default {
		    $this go$x $val
		    return
		}
	}
    }
    #
    proc menuNode {this x y} {
	catch {::destroy .menu$this}
	set m [menu .menu$this -tearoff 0]
	$m add command -label [mc Open] -command "
		$this select 0
		::destroy $m
	    "
	$m add command -label [mc Centre] -command "
		$this centre
		::destroy $m
	    "
	$m add command -label [mc Go] -command "
		$this go
		::destroy $m
	    "
	$m add separator
	$m add command -label [mc Stash] -command "[$this map] stashNode $this"
	set st normal
	switch [[$this map] stash]\
	  $this -\
	  {} { set st disabled }
	$m add cascade -label [mc Link] -menu [menu $m.links -tearoff 0]\
	  -state $st
	foreach {dir name} [array get ::[$this map]::linknames] {
	    $m.links add command -label $name -command "
		    $this linkSel $dir
		    ::destroy $m
		"
	}
	$m add separator
	$m add command -label [mc Unlink] -command "
		$this unlink 0
		::destroy $m
	    "
	$m add command -label [mc Relink] -command "
		$this relink 0
		::destroy $m
	    "
	$m add command -label [mc {Unlink Tree}] -command "
		$this unlink 1
		::destroy $m
	    "
	$m add command -label [mc {Relink Tree}] -command "
		$this relink 1
		::destroy $m
	    "
	$m add separator
	$m add command -label [mc Destroy] -command "
		$this destroy 0
		::destroy $m
	    "
	$m add command -label [mc {Destroy Tree}] -command "
		$this destroy 1
		::destroy $m
	    "
	tk_popup $m $x $y
	raise $m
    }
    #
    proc redraw {this} { [$this map] changeview }
    #
    proc unlink {this tree} {
	switch $this\
	  [[$this map] start] {
		tellError "Sorry, at the moment you cannot unlink the start\
		  node."
		return
	    }
	set ::${this}::deleted 1
	set chrt [$this map]
	foreach dir [[$this map] linkTypes] {
	    foreach lk [$this links($dir)] { $lk unlink $this $tree }
	}
	if {[winfo exists [set w [$chrt canvas].$this]]} {
	    if {[$chrt showdeleted]} {
		$w configure -background [set bg [tkDarken [$this background]\
		  200]]
		$w.lb configure -foreground [tkDarken [$this foreground] 200]\
		  -background $bg
		if {[listmember [$chrt selection] $this]} {
		    grid [$chrt window].info.f1.del
		}
	    } {
		::destroy $w
		[$chrt canvas] delete $this h$this
	    }
	}
	$this checkRooting
	if {![$chrt showdeleted]} { $chrt selectrem $this }
    }
    #
    proc relink {this tree} {
	set ::${this}::deleted 0
	foreach dir [[$this map] linkTypes] {
	    foreach lk [$this links($dir)] { $lk relink $this $tree }
	}
	if {[winfo exists [set w [[$this map] canvas].$this]]} {
	    $w configure -background [$this background]
	    $w.lb configure -foreground [$this foreground] -background\
	      [$this background]
	}
	$this checkRooting
    }
    #
    proc destroy {this tree} {
	foreach dir [[$this map] linkTypes] {
	    foreach lk [$this links($dir)] { $lk destroy $this $tree }
	}
	$this delete
    }
    #
    proc rootless {this} {
	foreach x [$this links(n)] { if {![$x deleted]} { return 0 } }
	return 1
    }
    #
    proc checkRooting {this} {
	if {![$this deleted] && [$this rootless]} {
	    [$this map] addRoot $this
	} {
	    [$this map] removeRoot $this
	}
    }
    #
    proc removeLink {this link} {
	foreach dir [[$this map] linkTypes] {
	    if {[listdel ::${this}::links($dir) $link]} break
	}
	$this checkRooting
    }
    #
    proc deLink {this link type} {
	listdel ::${this}::links($type) $link
	$this checkRooting
    }
    #
    proc urlimage {this} {
	return [string match http:* [set ::${this}::image]]
    }
    #
    proc html {this fd level tree} {
	set txt [::clean [string trim [$this notes]]]
	regsub -all "\n+" $txt </P><P> txt
	switch -- $level {
	1 {
		if {![$this urlimage]} {
		    switch {}\
		      [$this url] {
			    puts $fd <h1>[$this label]</h1>
			}\
		      default {
			    puts $fd "<h1><a href=\"[$this url]\">[$this\
			      label]</a></h1>"
			}
		} {
		    switch {}\
		      [$this url] {
			    puts $fd "<img src=\"[$this image]\" alt=\"[$this\
			      label]\">"
			}\
		      default {
			    puts $fd "<a href=\"[$this url]\"><img\
			      src=\"[$this image]\" border =\"0\" alt=\"[$this\
			      label]\"></a>"
			}

		}
		switch {}\
		  $txt {  }\
		  default {
			puts $fd <p>$txt</p>
		    }
		puts $fd <dl>
		foreach lk [$this links(s)] {
		    [$lk otherEnd $this] html $fd [expr {$level + 1}] $tree
		}
		puts $fd </dl>
	    }
	2 {
		if {![$this urlimage]} {
		    switch {}\
		      [$this url] {
			    puts $fd <dt>[$this label]</dt>
			}\
		      default {
			    puts $fd "<dt><a href=\"[$this url]\">[$this\
			      label]</a></dt>"
			}
		} {
		    switch {}\
		      [$this url] {
			    puts $fd "<dt><img src=\"[$this image]\"\
			      alt=\"[$this label]\"></dt>"
			}\
		      default {
			    puts $fd "<dt><a href=\"[$this url]\"><img\
			      src=\"[$this image]\" border =\"0\" alt=\"[$this\
			      label]\"></a></dt>"
			}
		}
		switch {}\
		  $txt {  }\
		  default {
			puts $fd <dd><p>$txt</p></dd>
		    }
	    }
	default {  }
	}
    }
    #
    proc text {this fd level tree} {
	set txt [string trim [$this notes]]
	switch -- $level {
	1 {
		puts $fd "[$this label]\n"
		switch {}\
		  $txt {  }\
		  default {
			puts $fd $txt
		    }
		puts $fd {}
		foreach lk [$this links(s)] {
		    [$lk otherEnd $this] html $fd [expr {$level + 1}] $tree
		}
	    }
	2 {
		puts $fd "[$this label]\n"
		switch {}\
		  $txt {  }\
		  default {
			puts $fd $txt
		    }
	    }
	default {  }
	}
    }
    #
    proc dot {this fd} {
        global ndone
	if {[info exists ndone($this)]} return
	set ndone($this) 1
	puts $fd "[$this name][$this dotprops];"
        foreach dir {s e w} {
	    foreach lk [$this links($dir)] {
	        set ot [$lk otherEnd $this]
	        puts $fd "[$this name] -> [$ot name];"
		$ot dot $fd
	    }
	}
    }
    #
    proc dotprops {this} {
       regsub -all "\"" [$this label] "\\\"" lb
       set props "label=\"$lb\""
#       append props {,shape=box}
#       append props ",fontcolour=[$this foreground]"
#       array set ftdata [font actual [$this font]]
#       append props ",fontsize=$ftdata(-size),fontname=\"$ftdata(-family)\""
       return "\[$props\]"
    }
    #
    proc addUserField {this pars str} { addUF $this $pars $str }
    #
    # ***** End of namespace node
}
#
proc mkLink {cv x y} {
    set item [$cv find closest $x $y]
    foreach t [$cv gettags $item] {
	switch -glob -- $t\
	  nod* {
		set node $t
	    }\
	  hnod* {
		set node [string range $t 1 end]
	    }\
	  current -\
	  sel -\
	  {[0-9]*} { }\
	  default {
		set type $t
	    }
    }
    global sx sy
    set chrt [$node map]
    upvar #0 ::${chrt}::ink ink ::${chrt}::thick thick
    foreach {x1 y1 x2 y2} [$cv bbox $node] break
    switch $type {
    s {
	    set sx [halfit $x1 $x2]
	    set sy $y2
	}
    w {
	    set sy [halfit $y1 $y2]
	    set sx $x1
	}
    e {
	    set sy [halfit $y1 $y2]
	    set sx $x2
	}
    n {
	    set sx [halfit $x1 $x2]
	    set sy $y1
	}
    }
    $cv create line $sx $sy [$cv canvasx $x] [$cv canvasy $y] -fill $ink\
      -tag rubber
    grab $cv
    bind $cv <Motion> "
	set cx \[$cv canvasx %x\]
	set cy \[$cv canvasy %y\]
	$cv coords rubber \$sx \$sy \$cx \$cy
    "
    bind $cv <ButtonRelease-1> "$node endLink $type $cv %x %y"
}
#
proc confNotes {nb chart} {
    set win [::Notebook::getNote $nb Notes]
    text $win.txt -width 30 -height 19 -yscrollcommand "$win.vs set" -wrap word
    trapid $win.txt $chart
    scrollbar $win.vs -command "$win.txt yview"
    grid $win.txt -row 0 -column 0 -sticky nsew
    grid $win.vs -row 0 -column 1 -stick ns
    grid columnconfigure $win 0 -weight 1
    grid rowconfigure $win 0 -weight 1
}
#
proc confFields {nb chart} {
    set win [::Notebook::getNote $nb Fields]
    set rw -1
    foreach {x b} {Image 1 URL 0 Mail 0 Chart 1 Object 1} {
	incr rw
	grid [label $win.l$rw -text [mc $x]] -row $rw -sticky w -padx 5
	incr rw
	grid [frame $win.$rw] -row $rw -sticky ew
	grid columnconfigure $win.$rw 0 -weight 1
	grid [entry $win.$rw.e] -sticky ew -padx 5
	if {$b} {
	    button $win.$rw.b -width 7 -text [mc Browse]
	    grid $win.$rw.b -row 0 -column 1
	}
	switch Image\
	  $x {
		bind $win.$rw.e <Return> "$chart changeview"
	    }
    }
    incr rw
    grid [button $win.uf -text [mc {User Defined Fields}]]
    filler $win [incr rw]
}
#
proc filler {win rw} {
    grid [frame $win.fill$rw] -row $rw -column 0 -sticky nsew
    grid rowconfigure $win $rw -weight 1
    grid columnconfigure $win 0 -weight 1
}
#
proc confFormat {nb chart} {
    set win [::Notebook::getNote $nb Format]
    grid [set sf [frame $win.st]] -sticky ew -row 0 -column 0 -pady 5
    grid [label $sf.sl -text [mc Style]] -row 0 -column 0 -sticky w -padx 5
    grid [menubutton $sf.mb -textvariable ::${chart}::sname -indicatoron 1\
      -menu $sf.mb.menu -relief raised -bd 2 -highlightthickness 2 -anchor c\
      -direction flush -width 12] -row 0 -column 1 -sticky ew
    menu $sf.mb.menu -tearoff 0
    $sf.mb.menu add command -label New -command "$chart newStyle $nb"
    $sf.mb.menu add separator
    foreach i [$chart styles] {
	$sf.mb.menu add -label [set ${i}::name] -command "$chart setStyle $nb\
	  $i"
    }
    grid columnconfigure $sf 1 -weight 1
    grid [frame $win.sep -bg black -borderwidth 4] -sticky ew -pady 5 -row 1\
      -column 0 -columnspan 2
    set rw 1
    foreach x {font foreground background} {
	incr rw
	grid [label $win.l$x -text [mc [capitalise $x]]] -row $rw -sticky w\
	  -padx 5
	incr rw
	grid [frame $win.$x] -row $rw -sticky ew
	grid columnconfigure $win.$x 0 -weight 1
	grid [entry $win.$x.e] -sticky ew -padx 5
	button $win.$x.b -width 7 -text [mc Select]
	grid $win.$x.b -row 0 -column 1
    }
    filler $win [incr rw]
}
#
proc confKeywords {nb chart} {
    set win [::Notebook::getNote $nb Keywords]
    grid [listbox $win.lb] -sticky nsew
    bind $win.lb <Enter> {focus %W}
    grid [frame $win.f] -sticky ew
    grid [label $win.f.new -text [mc New]:] -row 0 -column 0 -sticky w
    grid [entry $win.f.ent] -row 0 -column 1 -sticky ew
    grid columnconfigure $win.f 1 -weight 1
    grid rowconfigure $win.f 0 -weight 1
    grid rowconfigure $win 0 -weight 1
    grid columnconfigure $win 0 -weight 1
}
#
proc confAccess {nb chart} {
    global prog access
    set win [::Notebook::getNote $nb Access]
    grid [radiobutton $win.pub -text [mc Public] -variable access($chart)\
      -command "$chart access public"] -sticky w
    grid [radiobutton $win.prv -text [mc Private] -variable access($chart)\
      -command "$chart access private"] -sticky w
    grid [radiobutton $win.oth -text [mc Other] -variable access($chart)\
      -command "$chart access other"] -sticky w
    grid [listbox $win.lb] -sticky nsew
    bind $win.lb <Enter> {focus %W}
    grid [frame $win.f] -sticky ew
    grid [label $win.f.new -text [mc New]:] -row 0 -column 0 -sticky w
    set prog(fgdef) [$win.f.new cget -foreground]
    $win.f.new configure -foreground grey70
    grid [entry $win.f.ent -state disabled] -row 0 -column 1 -sticky ew
    grid columnconfigure $win.f 1 -weight 1
    grid rowconfigure $win.f 0 -weight 1
    grid rowconfigure $win 0 -weight 1
    grid columnconfigure $win 0 -weight 1
}
#
proc imageBrowse {node} {
    switch {}\
      [set file [tk_getOpenFile -defaultextension gif -title "Image for\
      [$node name]"]] return
    set ::${node}::image file:$file
}
#
proc chartBrowse {node} {
    switch {}\
      [set file [tk_getOpenFile -defaultextension cht -title "Chart link for\
      [$node name]"]] return
    set ::${node}::chart file:$file
}
#
proc objectBrowse {node} {
    switch {}\
      [set file [tk_getOpenFile -title "Object link for [$node name]"]] return
    set ::${node}::object file:$file
}
#
proc fileBrowse {var} {
    switch {}\
      [set file [tk_getOpenFile -title [mc "Select File"]]] return
    set $var file:$file
}



