#
# $Source: /home/nlfm/Working/Mercator/Released/lib/RCS/Util.tcl,v $
# $Date: 1999/12/15 13:24:38 $
# $Revision: 1.1.1.22 $
#
# ----------------------------------------------------------------------
#   AUTHOR:  Lindsay Marshall <lindsay.marshall@newcastle.ac.uk>
# ----------------------------------------------------------------------
# Copyright 1998 Lindsay F. Marshall (see COPYRIGHT)
# ======================================================================
#
package provide mercator 1.0
#
package require Object
package require http
package require configuration
package require Comms
package require Style
if {[catch {package require msgcat}]} {
    proc mc {string args} { return $string }
} {
    proc mc {string args} {
        return [::msgcat::mc $string]
    }
}
#
set GoodTypes [list image/gif image/ppm image/pgm]
set ImageTypes [list gif ppm pgm]                                               
if {![catch {package require Img}]} {
    set extensions(Img) 1
    lappend GoodTypes image/jpeg image/tiff image/bmp image/xpm image/png
    lappend ImageTypes jpeg tiff bmp xpm png
}                                       #
#
proc tellError {msg} {
    tk_messageBox -type ok -icon error -message $msg -title Error
}
#
proc tellInfo {msg} {
    tk_messageBox -type ok -icon info -message $msg -title Information
}
#
proc unimp {} {
    tk_messageBox -type ok -icon info -message "Sorry, not yet implemented." -title Sorry!
}
#
proc placeAd {} {
    global prog
    image create photo advert -file [file join $prog(lib) images $prog(program).gif]
    toplevel .ad
    set prog(fg) [[label .ad.f] cget -foreground]
    destroy .ad.f
    set w [winfo screenwidth .]
    set h [winfo screenheight .]
    wm overrideredirect .ad 1
    set iw [expr {[image width advert] + 10}]
    set ih [expr {[image height advert] + 10}]
    wm geometry .ad ${iw}x${ih}+[expr {($w -$ih) /2}]+[expr {($h - $ih )/2}]
    grid [label .ad.l -image advert -borderwidth 5 -relief ridge]
    bind .ad.l <1> {destroy .ad}
    update
}
#
proc splashScreen {} {
    placeAd
    after 1000 {catch {destroy .ad}}
}
#
proc newname {txt} {
    global nameCount
    return [string tolower $txt][incr nameCount]
}
#
proc Initialise {} {
    global Pmenus prog tcl_platform LinkTypes LinkPairs charts
#
    ::Comms::init
    ::Style::init
#
    package require $tcl_platform(platform)
#    
    platformInit
#
    array set LinkTypes {n Up s Down e Right w Left}
    array set LinkPairs {n s s n e w w e}
    array set Pmenus "
        Chart {
            {Open $prog(cmdkey)-o}
            {New $prog(cmdkey)-n}
            {Close $prog(cmdkey)-w}
            {Save $prog(cmdkey)-s -}
            {{Save As}}
            {-}
            {Print $prog(cmdkey)-p}
	    {> Export {
	    	    {HTML}
		    {Text}
		    {Dot}
	        }
	    }
            {-}
            {Quit $prog(cmdkey)-q}
        }
        Edit {
            {Undo {} -}
            {Cut {} -}
            {Copy {} -}
            {Paste {} -}
            {Clear {} -}
            -
            {{Select All} {$prog(cmdkey)-a} -}
	    -
	    {Find {$prog(cmdkey)-f}}
	    {{Find Again} {$prog(cmdkey)-g}}
	    -
	    {. {Fast Add} {fastadd}}
	    -
	    {{Global Preferences} {}}
	    {{Chart Preferences} {}}
        }
    	View {
		{@ Centred {view centred}}
		{@ Tree {view tree}}
		{@ Outline {view outline}}
		-
		{{Goto Stash} {} -}
		-
		{. Log tracking}
		{. {Node Info} nodeinfo}
		-
		{. {Show images} showimages}
		-
		{. {Show deleted} showdeleted}
		{. {Show All Links} showalllinks}
		-
	}
	Roots {
	}
	Windows {
	}
	Help {
	    {! {$prog(Program) $prog(version).$prog(patchlevel)}}
	    -
	    -
	    {Debug}
        }
    "
#
    wm withdraw .
    tk appname $prog(Program)
    option add *background #d0ccb8
#
# Hack to eliminate grab errors which cant be dealt with.
#
    if {![catch {rename grab sys_grab}]} {
	proc grab {args} {if {![catch {eval sys_grab $args} res]} { return $res } }
    }
#
#   Display the advert screen.
#
    splashScreen
#
    global nameCount base history prog recent ucount faces
#
#   Initialise the model class data structures.
#
    InitClasses
#
    set ucount 0
    set nameCount 0
    set recent {}
    set history {}
    set charts {}
    set prog(home) [lindex [glob ~] 0]
    if {[file exists [set hdir [prefdir]]]} {
	if {[file exists [file join $hdir preferences]]} {
            if {[catch {source [file join $hdir preferences]} msg]} {
	       die "Error reading your preferences file : $msg"
	    }
        }
       if {[file exists [file join $hdir recent]]} {
            if {[catch {source [file join $hdir recent]} msg]} {
	       die "Error reading your recent access data : $msg"
	    }
        }
    } {
        if {[catch {file mkdir $hdir} msg ]} {
	    die "Cannot create $prog(Program) directory : $msg"
        }
     }
}
#
proc makeMenu {obj focus var names} {
    upvar #0 $var menus
    set w [$obj window]
    foreach x $names {
        set lx [string tolower $x]
        $w.mb add cascade -label [mc $x] -menu $w.mb.$lx
        buildMenu $w.mb.$lx $obj $menus($x) $focus
    }
}
#
proc buildMenu {w obj stuff focus} {
    menu $w -tearoff 0
    set ccnt 0
    foreach x $stuff {
	foreach {typ label etc} $x break
	switch -- $typ {
	-   {$w add separator}
	!   {$w add command -label $label}
	.   { $w add checkbutton -label [mc $label] -variable ::${obj}::$etc }
	@   { $w add radiobutton -label [mc $label] \
		-variable ::${obj}::[lindex $etc 0] \
		-value [lindex $etc 1] }
	>   {
		$w add cascade -label [mc $label] -menu [set m $w.menu[incr ccnt]]
	        buildMenu $m $obj $etc $focus
	    }
	default {
		regsub -all { } $typ {} cmd
		$w add command -label [mc $typ] -command "$cmd $obj"
		switch {} $label {} default {
		    $w entryconfigure last -accelerator $label
		    bind $focus <$label> "$cmd $obj"
		}
		switch -- $etc - {
		    $w entryconfigure last -state disabled
		}
	    }
	}
    }
}
#
proc labent {w ix text var} {
    frame $w.f$ix
    grid columnconfigure $w.f$ix 1 -weight 1
    grid rowconfigure $w.f$ix 0 -weight 1
    label $w.f$ix.l -width 10 -text [mc $text] -anchor w
    entry $w.f$ix.e -textvariable $var
    grid $w.f$ix.l -row 0 -column 0 -sticky w
    grid $w.f$ix.e -row 0 -column 1 -sticky ew
    grid $w.f$ix -row $ix -sticky nsew
}
#
if {[catch {string title foo}]} {
proc capitalise {str} {
    return [string toupper [string index $str 0]][string range $str 1 end]
}
} {
proc capitalise {str} { return [string title $str] }
}
#
proc GotoStash {obj} { $obj gotoStash }
#
proc Quit {args} {
    exit 0
}
#
proc Open {obj} {
    global prog
    set types "
    {{[mc "$prog(Program) Files"]} .cht}
    {{[mc {All Files}]} *}
    "
    switch {} [set fn [tk_getOpenFile -defaultextension .cht \
      -filetypes $types -title [mc File]]] {
        return
    }
    [Chart -name [basename $fn] -file $fn] show
}
#
proc Close {obj} { $obj close }
#
proc Save {obj} { $obj save 0}
#
proc HTML {obj} {
    $obj exporthtml
}
#
proc Text {obj} {
    $obj exporttext
}
#
proc Dot {obj} {
    $obj exportdot
}
#
proc SaveAs {obj} { $obj save 1 }
#
proc Undo {obj} {
    unimp
}
#
proc Cut {obj} {
    unimp
}
#
proc Copy {obj} {
    unimp
}
#
proc SelectAll {obj} {
    unimp
}
#
proc Clear {obj} {
    unimp
}
#
proc Paste {obj} {
    unimp
}
#
proc Find {obj} {
    $obj searchInit
}
#
proc FindAgain {obj} {
    $obj searchAgain
}
#
proc Print {obj} { $obj print }
#
proc declare {type this pargs init} {
    proc $this {op args} "objcall [string tolower $type] $this \$op \$args"
    upvar #0 $this mdata
    array set mdata $init
    eval $this configure $pargs
    return $this
}
#
proc New {args} {
    global ucount
    [Chart -name Untitled[incr ucount]] show
}
#
proc trunctxt {obj args} {
    if {[string length [set txt [$obj label]]] > 12} {
        set ::${obj}::shorttext [string range $txt 0 9]...
    } {
        set ::${obj}::shorttext $txt
    }
}
#
proc Debug {args} {
    if {[winfo exists .debug]} {
        wm deiconify .debug
	raise .debug
	return
    }
    global DBGPos
    set DBGPos end
    toplevel [set f .debug]
    scrollbar $f.vs -command ".debug.txt yview"
    text $f.txt -yscrollcommand ".debug.vs set" -width 50 -height 20
    $f.txt tag configure input -foreground blue
    $f.txt tag configure output -foreground black
    $f.txt tag configure err -foreground red
    entry $f.ent
    bind $f.ent <Return> zDbgDo
    bind $f.ent <Control-p> "zDbgHist 2 $f.txt %W"
    bind $f.ent <Up> [bind $f.ent <Control-p>]
    bind $f.ent <Control-n> "zDbgHist -2 $f.txt %W"
    bind $f.ent <Down> [bind $f.ent <Control-n>]
    bind $f <Enter> "focus $f.ent"
    grid .debug.txt -row 0 -column 0 -sticky nsew
    grid .debug.vs -row 0 -column 1 -sticky ns
    grid .debug.ent -row 1 -column 0 -sticky ew
    grid columnconfigure .debug 0 -weight 1
    grid rowconfigure .debug 0 -weight 1
    .debug.txt configure -state disabled
}
#
proc basename {file} {
    return [file rootname [file tail $file]]
}
#
proc halfit {x1 x2} {
    return [expr {$x1 +($x2- $x1)/2}]
}
#
proc overlap {cv height width th tw x1 y1 ang} {
    set cang 0
    while {1} {
    	set xstep [expr {round($th*cos($ang))}]
	set ystep [expr {-round($th*sin($ang))}]
	set yswap 0
	set xswap 0
	while {1} {
	    switch {} [$cv find overlapping $x1 $y1 [expr {$x1+$tw}] \
	      [expr {$y1+$th}]]  { return [list $x1 $y1 $ang] }
	    if {[incr y1 $ystep] < 0 || ($y1+$th) > $height} {
	        set ystep [expr {-$ystep}]
	        incr y1 $ystep
		incr y1 $ystep
	        if {[incr yswap] > 1} break
	    }
	    if {[incr x1 $xstep] < 0 || ($x1+$tw) > $width} {
	        set xstep [expr {-$xstep}]
	        incr x1 $xstep
	        incr x1 $xstep
	        if {[incr xswap] > 1} break
	    }
	}
	set ang [expr {$ang - asin(1.0)/10.0}]
	if {[incr cang] > 9} { return [list $x1 $y1 $ang] }
    }
}
#
proc clean {txt} {
    regsub -all < $txt {\&lt;} txt
    regsub -all > $txt {\&gt;} txt
    regsub -all \" $txt {\&quot;} txt
    regsub -all ' $txt {\&apos;} txt
    regsub -all & $txt {\&amp;} txt
    return $txt
}
#
proc unclean {txt} {
    regsub -all {&lt;} $txt < txt
    regsub -all {&gt;} $txt > txt
    regsub -all {&quot;} $txt \" txt
    regsub -all {&apos;} $txt ' txt
    regsub -all {&amp;} $txt {\&} txt
    return $txt
}
#
proc quote {txt} {
    return \"[clean $txt]\"
}
#
proc listdel {list item} {
    upvar 1 $list lst
    if {[set ix [lsearch $lst $item]] >= 0} {
        set lst [lreplace $lst $ix $ix]
	return 1
    }
    return 0
}
#
proc listmember {list item} {
    return [expr {[lsearch $list $item] >= 0}]
}
#
proc listadd {list item} {
    upvar 1 $list lst
    if {![listmember $lst $item]} { lappend lst $item ; return 1 }
    return 0
}
#
proc makeImage {file} {
    if {![catch {image width $file}]} { return $file }
    set del 0
    set fname $file
    switch -glob -- $file {
        http:* {
	    set token [::http::geturl $file]
	    ::http::wait $token
	    switch [::http::status $token] ok {
	        global GoodTypes
	        upvar #0 $token state
		array set meta $state(meta)
		switch -glob -- [set typ [string trim $meta(Content-Type)]] image/* {
		    if {[lsearch -exact $GoodTypes $typ] < 0} {
	                tellError "Cannot render \"$file\" ($typ)\n\
			 Must be one of $GoodTypes"
		        return {}
		    }
		    regexp {.*/(.*)} $typ m ext
		} default {
		    tellError "\"$file\" is not an image"
		    return {}
		}
	        set file [file join [tmpdir] [pid].$ext]
		set fd [open $file w]
		fconfigure $fd -translation binary
		puts $fd [::http::data $token]
		close $fd
	    } default {
	        tellError "Can't access image \"$file\""
		return {}
	    }
	    set del 1
	}
        file:* {
	    regsub {^file:} $file {} file
	}
    }
    if {[catch {image create photo $fname -file $file} msg]} {
        tellError "Cannot create image from \"$file\" - $msg"
	if {$del} { file delete $file }
	return {}
    }
    if {$del} { file delete $file }
    return $fname
}
#
proc nsdump {nm} {
    foreach x [info vars ::${nm}::*] {
        if {[catch {puts stderr "$x :{[set $x]}"}]} {
	    parray $x
	}
    }
}
#
proc getval {ns var def} {
    if {![info exists ${ns}::$var]} return $def
    return [set ${ns}::$var]
}
#
proc saveData {this desc} {
    foreach {x prot} {image {} url {} mail mailto: chart {}} {
	switch {} [$this $x] continue
	puts $desc "<object name=[quote $x] href=[quote ${prot}[clean [$this $x]]]/>"
    }
    foreach x [namespace children ::${this}] {
	puts -nonewline $desc " <field name=[quote [namespace tail $x]]"
	foreach y [namespace eval $x {info vars UF_*}] {
	    puts -nonewline $desc \
	      " [string range $y 3 end]=[quote [set ${x}::$y]]"
	}
	puts $desc >
	puts $desc [clean [set ${x}::value]]
	puts $desc </field>
    }
    switch {} [set ::${this}::notes] {} default {
	puts $desc <notes>
	puts $desc [clean [set ::${this}::notes]]
	puts $desc </notes>
    }
}
#
proc addUF {this pars str} {
    array set values $pars
    if {![info exists values(name)]} {
	set values(name) [newname uf]
    }
    foreach {x y} [array get values] {
	switch name $x continue
	namespace eval ::${this}::$values(name) "variable UF_$x [list [unclean $y]]"
    }
    regsub ^\n* $str {} str
    regsub {\n*$} $str {} str
    namespace eval ::${this}::$values(name) "variable value [list [unclean $str]]"
}
#
proc trapid {win chart} {
    uplevel #0 rename $win real_$win
    uplevel #0 proc $win {{sub args}} "{
       switch -- \$sub insert - delete { $chart used }
       eval real_$win \$sub \$args
    }"
    bind $win <Destroy> " rename $win {} "
}
